import keysightSD1
import keysight_hvi as kthvi

"""
Pathwave Test Sync Executive Multi-Channel Sync Playback

Copyright Keysight Technologies 2022

In this programming example, Pathwave Sync Executive is used to program multiple M3xxxA AWGs to synchronously output first a FP (Front Panel) trigger
pulse and afterwards a previously queued waveform. All modules run fully synchronized and actions across modules can be controlled with the 
timing resolution of the M3xxxA AWGs which is of 10ns.

The showcased HVI API functionalities are:
* Scalability to an arbitrary number of instruments. In this specific case M3xxxA PXI AWGs
* Off-shelf inter-instrument synchronization capabilities
* HVI Synchronized Multi-Sequence Block to synchronously parallel-execute sequences in multiple instruments
* HVI core and HVI product-specific instructions

The Python code that follows: 
1. Opens an arbitrary number of M3xxxA AWGs (real hardware or simulation mode) 
2. Defines HVI resources (engines, chassis, chassis interconnections, trigger, action, events, HVI triggering resources, etc)
3. Programs HVI sequences by adding the necessary statements and instructions in each AWG's sequence
4. Compiles HVI, loads it to HW and executes it

For further info please check the programming example titled "Pathwave Test Sync Executive Multi-Channel Sync Playback using Keysight M3xxxA Arbitrary Waveform Generators"
"""

######################################
# Application Configuration 
######################################

class ApplicationConfig:
    """
    Defines module descriptors, configuration options and names of HVI engines, actions, triggers
    """
    def __init__(self):
        """
        Define PXI instruments options
        """
        # Define options to open the instruments. 
        # Complete option list can be found in the SD1 3.x User Guide
        self.options = 'channelNumbering=keysight'
        # Simulated HW Option
        self.hardware_simulated = True
        """ 
        Define names of HVI engines, actions, triggers
        """
        # Define names of HVI engines, actions, triggers
        self.engine_name = "AwgEngine"
        self.awg_trigger_action_name = "AwgTrigger"
        self.fp_trigger_name = "FpTrigger"
        """
        Define HW Platform
        """
        # Define module descriptors below with your instruments information
        self.module_descriptors = [
            ModuleDescriptor('M3202A', 1, 4, self.options, ""),
            ModuleDescriptor('M3202A', 1, 5, self.options, "")]
        self.define_engine_names()
        
        # Define list of chassis containing instruments to be used by HVI
        self.chassis_list = [1]
        
        # Multi-chassis setup
        # Define the System Sync Modules included in your system.
        # For example: 
        # self.system_sync_module_descriptors = [
        #    SystemSyncModuleDescriptor('PXI0::CHASSIS2::SLOT10::INDEX0::INSTR', self.ssm_options)]
        self.ssm_options = ''
        self.ssm_simulation_options = 'Simulate=true,DriverSetup=Model=M9033A'
        self.system_sync_modules_descriptors = []
        # For each SSM define which SSM is connected to its downstream connectors.
        # For example: 
        # self.ssm_connections = [
        #      SystemSyncModuleConnection(ssm1_chassis=1, ssm1_downstream_connector_number=3, ssm2_chassis=2)]
        self.ssm_connections = []

        # Assign triggers to HVI object to be used for HVI-managed synchronization, data sharing, etc
        # NOTE: In a multi-chassis setup ALL the PXI lines listed below need to be shared among each M9031 board pair by means of SMB cable connections
        self.pxi_sync_trigger_resources = [
            kthvi.TriggerResourceId.PXI_TRIGGER0,
            kthvi.TriggerResourceId.PXI_TRIGGER1]
        """
        AWG Parameters
        """
        # AWG settings for all channels
        self.sync_mode = keysightSD1.SD_SyncModes.SYNC_NONE
        self.queue_mode = keysightSD1.SD_QueueMode.ONE_SHOT
        self.start_delay = 0
        self.prescaler = 0
        self.num_cycles = 0
        self.amplitude = 1
        # Trigger settings
        self.trigger_mode = keysightSD1.SD_TriggerModes.SWHVITRIG_CYCLE

    def define_engine_names(self):
        " Define names for all HVI engines"
        for index in range(len(self.module_descriptors)): # module descriptor list populate dn application config class
            self.module_descriptors[index].engine_name = self.engine_name + str(index)

######################################
# Main Program
######################################

def main():
    ###################################################
    # Open PXI modules, setup AWGs and queue waveforms
    ###################################################
    # Load configuration
    config = ApplicationConfig()

    # Checks flags defining if the code executes on HW or Simulation mode
    if config.hardware_simulated:
        print("Code running in Simulation Mode")

    # Open instruments
    module_dict = open_modules(config)
    
    # Queue AWG Waveforms
    for module in module_dict.values():
        configure_awg(config, module.instrument, module.num_channels)

    ######################################
    # System definition
    ######################################
    # Create system definition object
    my_system = kthvi.SystemDefinition("MySystem") 
    # Define your system, HW platform, add HVI resources
    define_hvi_resources(my_system, module_dict, config)
    # Output to console the details of the system definition
    output_sys_def(my_system, config.module_descriptors)
    
    ######################################
    # Program HVI sequences
    ######################################
    print("-----------------------")
    print("Program HVI Sequences")
    print("-----------------------")
    print("Initializing the defined system...")
    # Create sequencer object
    sequencer = kthvi.Sequencer("MySequencer", my_system)
    print("Programming the HVI sequences...")
    # Program HVI sequences using the sequencer object
    program_mimo_trigger_sequence(sequencer, module_dict, config)
    # Export the programmed sequence to text
    export_hvi_sequences(sequencer)
    
    ########################################
    # Compile, Load to HW, Run HVI Sequence
    ########################################
    print("-------------------")
    print("Execute HVI")
    print("-------------------")
    print("Compiling HVI...")
    # Compile HVI sequences
    hvi = sequencer.compile()
    print(hvi.compile_status.to_string())
    print("HVI Compiled")
    print("This HVI needs to reserve {} PXI trigger resources to execute".format(len(hvi.compile_status.sync_resources)))
    
    # Load HVI to HW: load sequences, configure actions/triggers/events, lock resources, etc.
    hvi.load_to_hw()    
    print("HVI Loaded to HW")

    # Execute HVI in blocking mode: SW waits until HVI sequences ends their execution
    # Eventually enter a timeout for the HVI execution to be stopped: timeout = timedelta(seconds=0), hvi.run(timeout)
    hvi.run(hvi.no_timeout)
    print("HVI Running...")

    # User-controlled HVI execution
    if config.hardware_simulated:
        print("Simulation completed successfully")
    else:
        interact_with_hvi(hvi)

    # Release HW resources once HVI execution is completed
    hvi.release_hw()
    print("Releasing HW...")

    # Close all modules at the end of the execution
    for engine_name in module_dict:
        module_dict[engine_name].instrument.close()
    print("PXI modules closed\n")

########################################################
# 2nd Level: Functions to Define, Program, Execute HVI
#######################################################

def define_hvi_resources(sys_def, module_dict, config):
    """
    Configures all the necessary resources for the HVI application to execute: HW platform, engines, actions, triggers, etc. 
    """
    # Define HW platform: chassis, interconnections, PXI trigger resources, synchronization, HVI clocks
    define_hw_platform(sys_def, config)

    # Define all the HVI engines to be included in the HVI
    define_hvi_engines(sys_def, module_dict)
    
    # Define list of actions to be executed
    define_hvi_actions(sys_def, module_dict, config)

    # Defines the trigger resources
    define_hvi_triggers(sys_def, module_dict, config)


def define_hw_platform(sys_def, config):
    """
    Define HW platform: chassis, interconnections, PXI trigger resources, synchronization, HVI clocks 
    """
    # Define chassis resources 
    # For multi-chassis setup details see programming example documentation
    for chassis_number in config.chassis_list:
        if config.hardware_simulated:
            sys_def.chassis.add_with_options(chassis_number, 'Simulate=True,DriverSetup=Model=GenericPxieChassis')
        else:
            sys_def.chassis.add(chassis_number)

    # Define System Sync Modules (SSMs)
    if config.system_sync_modules_descriptors:
        interconnects = sys_def.interconnects
        ssm_list = []
        for descriptor in config.system_sync_modules_descriptors:
            if config.hardware_simulated:
                ssm = interconnects.add_sync_module(descriptor.resource_id, config.ssm_simulation_options)
            else:
                ssm = interconnects.add_sync_module(descriptor.resource_id, descriptor.options)
            # Append SSM object
            ssm_list.append(ssm)
        # Define connections between SSMs
        if config.ssm_connections:
            for connection in config.ssm_connections:
                connector_number = connection.ssm1_downstream_connector_number
                for ssm in ssm_list:
                    if ssm.chassis == connection.ssm1_chassis:
                        ssm1 = ssm
                    if ssm.chassis == connection.ssm2_chassis:
                        ssm2 = ssm
                # Implement each user-defined connection
                try:
                    # Set connection. SSMs have always one upstream port
                    ssm1.connectivity.systemsync_downstream[connector_number].set_connection(ssm2.connectivity.systemsync_upstream[0])
                except:
                    exit("Exception! Please check the valued defined for SyncModule resource ids, chassis numbers and connections")
    
    # Assign the defined PXI trigger resources
    sys_def.sync_resources = config.pxi_sync_trigger_resources
    
    # Assign clock frequencies that are outside the set of the clock frequencies of each HVI engine
    # Use the code line below if you want the application to be in sync with the 10 MHz clock
    sys_def.non_hvi_core_clocks = [10e6]


def define_hvi_engines(sys_def, module_dict):
    """
    Define all the HVI engines to be included in the HVI
    """
    # For each instrument to be used in the HVI application add its HVI Engine to the HVI Engine Collection 
    for engine_name in module_dict.keys():
        sys_def.engines.add(module_dict[engine_name].hvi.engines.main_engine, engine_name)


def define_hvi_actions(sys_def, module_dict, config):
    """ 
    Defines AWG trigger actions for each module, to be executed by the "action execute" instruction in the HVI sequence
    Create a list of AWG trigger actions for each AWG module. The list depends on the number of channels 
    """
    # For each AWG, define the list of HVI Actions to be executed and add such list to its own HVI Action Collection
    for engine_name in module_dict.keys():
        for ch_index in range(1, module_dict[engine_name].num_channels + 1):
            # Actions need to be added to the engine's action list so that they can be executed
            action_name = config.awg_trigger_action_name + str(ch_index) # arbitrary user-defined name
            instrument_action = "awg{}_trigger".format(ch_index) # name decided by instrument API
            action_id = getattr(module_dict[engine_name].hvi.actions, instrument_action)
            sys_def.engines[engine_name].actions.add(action_id, action_name)


def define_hvi_triggers(sys_def, module_dict, config):
    """
    Defines and configure the FP trigger output of each AWG 
    """
    # Add to the HVI Trigger Collection of each HVI Engine the FP Trigger object of that same instrument
    for engine_name in module_dict.keys():
        fp_trigger_id = module_dict[engine_name].hvi.triggers.front_panel_1
        fp_trigger = sys_def.engines[engine_name].triggers.add(fp_trigger_id, config.fp_trigger_name)
        # Configure FP trigger in each hvi.engines[index]
        fp_trigger.config.direction = kthvi.Direction.OUTPUT
        fp_trigger.config.polarity = kthvi.Polarity.ACTIVE_HIGH
        fp_trigger.config.sync_mode = kthvi.SyncMode.IMMEDIATE
        fp_trigger.config.hw_routing_delay = 0
        fp_trigger.config.trigger_mode = kthvi.TriggerMode.LEVEL 
        #NOTE: FP trigger pulse length is defined by the HVI Statements that control FP Trigger ON/OFF


def output_sys_def(sys_def, descriptors):
    """
    Outputs the details of each instrument and engine included in the system definition.
    """
    print("-------------------")
    print("System Definition")
    print("-------------------")
    # Output KS2201A HVI Core Version
    hvi_core_version = kthvi.SystemDefinition.hvi_core_version
    print("The PathWave Test Sync Executive API installed on your system has an HVI core version " + hvi_core_version.to_string() + "\n")
    # Output info about the HW instruments
    print("HW Instruments:")
    for descriptor in descriptors:
        engine = sys_def.engines[descriptor.engine_name]
        print("- Model: {} in chassis: {}, slot: {}, HVI Engine Name: {}, HVI core version: ".format(descriptor.model_number, descriptor.chassis_number, descriptor.slot_number, descriptor.engine_name) + 
        engine.software_version.to_string() + ", HVI Engine FPGA IP version: " + engine.firmware_version.to_string())
    # Output info about the Sync Modules
    if sys_def.interconnects:
        print("System Sync Modules:")
        for index in range(sys_def.interconnects.count):
            sync_module = sys_def.interconnects[index]
            len_upstream = len(sync_module.connectivity.systemsync_upstream)
            len_downstream = len(sync_module.connectivity.systemsync_downstream)
            print("- System Sync Module in chassis: {}, slot: {} with {} Upstream Ports and {} Downstream Ports".format(sync_module.chassis, sync_module.slot, len_upstream, len_downstream))


def program_mimo_trigger_sequence(sequencer, module_dict, config):
    """ 
    Programs the MIMO Trigger HVI sync sequence
    """
    # Add a Sync Multi-Sequence Block (SMSB)
    sync_block = sequencer.sync_sequence.add_sync_multi_sequence_block("TriggerAWGs", 30)
    # Program the SMSB to trigger AWGs
    program_trigger_awgs(sync_block, module_dict, config)


def program_trigger_awgs(sync_block, module_dict, config):
    " Programs the Sync Multi-Sequence Block to trigger AWGs "

    # Add instructions to each local sequence within the multi-sequence block
    for index in range(0, sync_block.sequences.count):
        
        # Retrieve the engine sequence from the collection of HVI Local Sequences
        # HVI Local Sequence Collection is automatically created form the user-defined HVI Engine Collection
        # Each HVI Local Sequence can be retrieved using the name alias of the corresponding HVI Engine
        sequence = sync_block.sequences[index]
        # Retrieve FP Trigger prefiously defined in the HVI Trigger Collection
        fp_trigger = sequence.engine.triggers[config.fp_trigger_name]
        # Retrive TriggerWrite instruction from HVI Native InstructionSet
        trigger_write = sequence.instruction_set.trigger_write
        
        # Write FP Trigger ON
        instr_trigger_ON = sequence.add_instruction("FP Trigger ON", 10, trigger_write.id)
        instr_trigger_ON.set_parameter(trigger_write.trigger.id, fp_trigger)
        instr_trigger_ON.set_parameter(trigger_write.sync_mode.id, trigger_write.sync_mode.immediate)
        instr_trigger_ON.set_parameter(trigger_write.value.id, trigger_write.value.on)

        # Write FP Trigger OFF
        instr_trigger_OFF = sequence.add_instruction("FP Trigger OFF", 100, trigger_write.id)
        instr_trigger_OFF.set_parameter(trigger_write.trigger.id, fp_trigger)
        instr_trigger_OFF.set_parameter(trigger_write.sync_mode.id, trigger_write.sync_mode.immediate)
        instr_trigger_OFF.set_parameter(trigger_write.value.id, trigger_write.value.off)

        # Execute AWG trigger from the HVI sequence of each module
        # "Action Execute" instruction executes the AWG trigger from HVI
        action_list = sequence.engine.actions
        instruction1 = sequence.add_instruction("AWG trigger", 10, sequence.instruction_set.action_execute.id)
        instruction1.set_parameter(sequence.instruction_set.action_execute.action.id, action_list)


def export_hvi_sequences(sequencer):
    """
    Exports the programmed HVI sequences to text format
    """
    # Generate HVI sequence description text
    output = sequencer.sync_sequence.to_string(kthvi.OutputFormat.DEBUG)
    
    # Write file
    text_file = open("TextOutput/HviMultiAwgSync.txt", "w")
    text_file.write(output)
    text_file.close()
    print("Programmed HVI sequences exported to file")


def interact_with_hvi(hvi):
    " User-controlled HVI execution "
    # If running with hardware, iteratively re-run HVI until the user decides to exit
    while True:
        print("Press enter to run HVI again, q to exit...")
        if input() == 'q':
            break
        hvi.run(hvi.no_timeout)

#################################################################
# 3rd Level: Classes and Functions to Use SD1/M3xxxA Instruments
#################################################################

class ModuleDescriptor:
    "Descriptor for module objects"
    def __init__(self, model_number, chassis_number, slot_number, options, engine_name):
        self.model_number = model_number
        self.chassis_number = chassis_number
        self.slot_number = slot_number
        self.options = options
        self.engine_name = engine_name


class SystemSyncModuleDescriptor:
    """
    Descriptor for System Sync Modules
    """
    def __init__(self, resource_id, options):
        self.resource_id = resource_id
        self.options = options


class SystemSyncModuleConnection:
    """
    Describes the connection between two SyncModule objects ssm1 and ssm2. SSMs have always one upstream port:
    ssm1.connectivity.systemsync_downstream[connector_number].set_connection(ssm2.connectivity.systemsync_upstream[0])
    """
    def __init__(self, ssm1_chassis, ssm1_downstream_connector_number, ssm2_chassis):
        self.ssm1_chassis = ssm1_chassis
        self.ssm1_downstream_connector_number = ssm1_downstream_connector_number
        self.ssm2_chassis = ssm2_chassis


class Module:
    "Class defining a modular instrument object and its properties"
    def __init__(self, instrument_object, hvi_interface, num_channels):
        self.instrument = instrument_object
        self.hvi = hvi_interface
        self.num_channels = num_channels


def open_modules(config):
    """ 
    Opens and creates each M3xxxA instrument object.
    Returns a dictionary of module objects whose keys are the HVI engine names.
    Please check SD1 3.x User Guide for options to open the instrument objects
    """
    # Initialize output variables
    num_modules = 0
    module_dict = {} # dictionary of modules

    # Open SD1 instrument objects
    for descriptor in config.module_descriptors:
        if descriptor.model_number == "M3102A":
            instr_obj = keysightSD1.SD_AIN()
        else:
            instr_obj = keysightSD1.SD_AOU()
        instr_obj_options = descriptor.options + ',simulate=true' if config.hardware_simulated else descriptor.options
        id = instr_obj.openWithOptions(descriptor.model_number, descriptor.chassis_number, descriptor.slot_number, instr_obj_options)
        if id < 0:
            raise Exception("Error opening instrument in chassis: {}, slot: {}! Error code: {}, {}. Exiting...".format(descriptor.chassis_number, descriptor.slot_number, id, keysightSD1.SD_Error.getErrorMessage(id)))
        nCh = instr_obj.getOptions("channels")
        if nCh == "CH2":
            num_channels = 2
        elif nCh == "CH4":
            num_channels = 4
        else:
            raise Exception("PXI module in chassis {}, slot {} returned number of channels = {} which is incorrect. Exiting... ".format(instr_obj.getChassis(), instr_obj.getSlot(), nCh))
        module_dict[descriptor.engine_name] = Module(instr_obj, instr_obj.hvi, num_channels)
        num_modules += 1
    
    return module_dict


def configure_awg(config, awg_module, num_channels = 2):
    """ 
    Queues a Gaussian waveform to all the num_channels channels of
    the AWG object awg_module that is passed to the function
    """  
    # Load waveform to AWG memory
    awg_module.waveformFlush() #memory flush
    wave = keysightSD1.SD_Wave()
    wfmNum = 0
    wave.newFromFile("waveforms/Gaussian.csv")
    awg_module.waveformLoad(wave, wfmNum)

    for channel in range(1, num_channels+1):
        # AWG queue flush 
        awg_module.AWGstop(channel)
        awg_module.AWGflush(channel)
        # Set AWG mode
        awg_module.channelWaveShape(channel, keysightSD1.SD_Waveshapes.AOU_AWG)
        awg_module.channelAmplitude(channel, config.amplitude)
        # AWG configuration
        awg_module.AWGqueueConfig(channel, config.queue_mode)
        awg_module.AWGqueueSyncMode(channel, config.sync_mode)
        # Queue waveform to the selected channel
        awg_module.AWGqueueWaveform(channel, wfmNum, config.trigger_mode, config.start_delay, config.num_cycles, config.prescaler)
        awg_module.AWGstart(channel) # AWG starts and waits for an AWG trigger


if __name__ == '__main__':
    main()